#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


from __future__ import print_function

import pprint

print("Importing child.")


def displayDict(d):
    d = dict(d)

    if "__loader__" in d:
        d["__loader__"] = "<__loader__ removed>"

    if "__file__" in d:
        d["__file__"] = "<__file__ removed>"

    if "__firstlineno__" in d:
        del d["__firstlineno__"]

    if "__static_attributes__" in d:
        del d["__static_attributes__"]

    # Avoid recursion that we don't offer for classes.
    if "__locals__" in d:
        del d["__locals__"]

    return pprint.pformat(d)


class A:
    pass


print("Class defined here, has these vars", displayDict(vars(A)))

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
