__version__="4.10.1"
__copyright__="Copyright (c) 2001 - 2025 The SCons Foundation"
__developer__="bdbaddog"
__date__="Sun, 16 Nov 2025 14:18:20 -0700"
__buildsys__="M1Dog2021"
__revision__="055b01f429d58b686701a56df863a817c36bb103"
__build__="055b01f429d58b686701a56df863a817c36bb103"
# make sure compatibility is always in place
import SCons.compat  # noqa